#ifndef _Bus_I2C_H
#define _Bus_I2C_H

#include "Prefix.h"
#include <vector>

namespace EARTH {

namespace FPGA {
	class I2C;
	class I2C_;
}

namespace Bus {
	class I2C {
	public:
		I2C()
		{
			mInstructionAddress = 0;
			mReadAddress = 0;
			mEnd = false;
			mInstructionCount = 0;
		}

		~I2C()
		{
		}

		void SetInstructionAddress(uint32 address)
		{
			mInstructionAddress = address;
		}

		uint32 InstructionAddress() const
		{
			return mInstructionAddress;
		}

		void Start()
		{
			AddInstruction(I_DATA_H);
			AddInstruction(I_CLOCK_H);
			AddInstruction(I_DATA_L);
			AddInstruction(I_CLOCK_L);
		}

		void Stop()
		{
//			AddInstruction(I_CLOCK_L);
			AddInstruction(I_DATA_L);
			AddInstruction(I_CLOCK_H);
			AddInstruction(I_DATA_H);
		}

		void Write(const uint8 *ptr, uint32 size = 1)
		{
			_ASSERT(ptr);
			_ASSERT(size);

			for (uint32 i=0; i<size; i++) {
				uint8	byte = ptr[i];
				for (uint32 j=0; j<8; j++) {
					AddInstruction(BIT_SHIFT_MASK(byte, 7-j, 1) ? I_DATA_H_NOP : I_DATA_L_NOP);
				}

				AddInstruction(I_DATA_H_ACK0);
			}
		}

		void Write(uint8 data)
		{
			Write(&data, sizeof(data));
		}

		uint32 ReadAddress() const
		{
			return mReadAddress;
		}

		uint32 Read(uint32 size = 1)
		{
			_ASSERT(size);

			for (uint32 i=0; i<size; i++) {
				for (uint32 j=0; j<8; j++) {
					AddInstruction(I_DATA_H_READ);
				}

				bool	end = (i == size-1);
				AddInstruction(end ? I_DATA_H_NOP : I_DATA_L_NOP);
			}

			uint32	readAddress = mReadAddress;
			mReadAddress += size;

			return readAddress;
		}

		void Reset()
		{
			AddInstruction(I_RESET);
		}

		void Sleep(uint32 ms)
		{
			for (uint32 i=0; i<ms; i++) {
				AddInstruction(I_SLEEP);
			}
		}

		uint32 InstructionCount() const
		{
			return mInstructionCount;
		}

		void End()
		{
			AddInstruction(I_END);
			mEnd = true;

			if (mInstructionCount % 2) {
				AddInstruction(I_END);
			}
		}

		bool IsEnd() const
		{
			return mEnd;
		}

		uint8 Data1(size_t index) const
		{
			return static_cast<uint8>(DataN(index, 1));
		}

		uint16 Data2(size_t index) const
		{
			return static_cast<uint16>(DataN(index, 2));
		}

		uint32 Data3(size_t index) const
		{
			return DataN(index, 3);
		}

		uint32 Data4(size_t index) const
		{
			return DataN(index, 4);
		}

	protected:
		friend class FPGA::I2C;
		friend class FPGA::I2C_;
		friend class I2C_;

		uint32				mInstructionAddress;

		uint32				mReadAddress;
		std::vector<uint8>	mReadData;
		bool				mEnd;

		uint32				mInstructionCount;
		uint32				mInstructionTemp;

		enum {
			L,
			H
		};

		enum {
			I_END,
			I_ADDRESS,
			I_CLOCK_L,
			I_CLOCK_H,
			I_DATA_L,
			I_DATA_H,
			I_RESET,
			I_SLEEP,	// Sleep 1ms
			I_DATA_L_NOP  = BIT_4(1,0,0,0),
			I_DATA_H_NOP  = BIT_4(1,1,0,0),
			I_DATA_H_READ = BIT_4(1,1,0,1),
			I_DATA_H_ACK0 = BIT_4(1,1,1,0),
			I_DATA_H_ACK1 = BIT_4(1,1,1,1),

			// eXgp
			_I_DATA_L_READ = I_DATA_H_READ ^ BIT_4(0,1,0,0)
//			_I_DATA_L_ACK0 = I_DATA_H_ACK0 ^ BIT_4(0,1,0,0),
//			_I_DATA_L_ACK1 = I_DATA_H_ACK1 ^ BIT_4(0,1,0,0)
		};

		std::vector<uint8>	mInstructions;

		void SetReadAddress(uint32 address)
		{
			AddInstruction(I_ADDRESS);
			AddInstruction(BIT_SHIFT_MASK(address, 4*(0), 4));
			AddInstruction(BIT_SHIFT_MASK(address, 4*(1), 4));
			AddInstruction(BIT_SHIFT_MASK(address, 4*(2), 3));

			mReadAddress = address;
		}

		void ClearReadData()
		{
			mReadData.clear();
		}

		void PushReadData(uint8 data)
		{
			mReadData.push_back(data);
		}

		void AddInstruction(uint32 instruction)
		{
			_ASSERT(mInstructionCount < 4*1024);

			if ((mInstructionCount % 2) == 0) {
				mInstructionTemp = instruction;
			} else {
				mInstructionTemp |= instruction << 4;
			}

			if (mInstructionCount % 2) {
				mInstructions.push_back(static_cast<uint8>(mInstructionTemp));
			}

			mInstructionCount++;
		}

		uint32 DataN(size_t index, uint32 n) const
		{
			_ASSERT((1 <= n) && (n <= 4));

			uint32 data = 0;

			for (uint32 i=0; i<n; i++) {
				data <<= 8;
				data  |= mReadData[index + i];
			}

			return data;
		}

	};
}
}

#endif
