#ifndef _QM1D1C0042_H
#define _QM1D1C0042_H

#include "Prefix.h"
#include "OS_Thread.h"
#include "Tuner_S.h"
#include "TC90522_S.h"
#include "OS_Time.h"

#define I2C_NULL NULL

namespace EARTH {
	class QM1D1C0042 : public Tuner_S {
	public:
		static uint8 Address(uint32 index)
		{
			const uint8		ADDRESS[2] = { 0x63, 0x60 };
			_ASSERT(index < sizeof(ADDRESS)/sizeof(*ADDRESS));
			return ADDRESS[index];
		}

		QM1D1C0042(TC90522_S *tc90522) : Tuner_S(tc90522)
		{
			_ASSERT(mTC90522);

			mSleep = false;

			mChannel = 0;
			mOffset  = 0;

			mWriteFunction = NULL;
			mReadFunction  = NULL;
			mSleepFunction = NULL;
		}

		~QM1D1C0042()
		{
		}

		typedef void (*WriteFunction)(uint8, uint8  , void *);
		typedef void (*ReadFunction )(uint8, uint8 *, void *);
		typedef void (*SleepFunction)(uint32        , void *);

		void SetCallbackFunctions(WriteFunction write, ReadFunction read, SleepFunction sleep, void *arg)
		{
			mWriteFunction = write;
			mReadFunction  = read;
			mSleepFunction = sleep;
			mFunctionArg   = arg;
		}

		struct Param {
			uint32	channelFrequencykHz;
			uint32	crystalFrequencykHz;
			bool	fastSearchMode;
			bool	standby;
			uint32	lowPassFilterWaitTime;
			uint32	fastSearchWaitTime;
			uint32	normalSearchWaitTime;
		};

		Param	*ParamPtr()
		{
			return &mParam;
		}

		void DummyReset(Bus::I2C *bus)
		{
			const uint8		INIT_DUMMY_RESET = 0x0C;

			/*status =*/Write(bus, 0x01, INIT_DUMMY_RESET);
			/*status =*/Write(bus, 0x01, INIT_DUMMY_RESET);
		}

		void InitRegParam()
		{
			const uint8 rw_reg[0x20] = {
				0x48,	// 0x00
				0x1c,	// 0x01
				0xa0,	// 0x02
				0x10,	// 0x03
				0xbc,	// 0x04
				0xc5,	// 0x05
				0x20,	// 0x06
				0x33,	// 0x07
				0x06,	// 0x08
				0x00,	// 0x09
				0x00,	// 0x0a
				0x00,	// 0x0b
				0x03,	// 0x0c
				0x00,	// 0x0d
				0x00,	// 0x0e
				0x00,	// 0x0f
				0x00,	// 0x10
				0xff,	// 0x11
				0xf3,	// 0x12
				0x00,	// 0x13
				0x2a,	// 0x14
				0x64,	// 0x15
				0xa6,	// 0x16
				0x86,	// 0x17
				0x8c,	// 0x18
				0xcf,	// 0x19
				0xb8,	// 0x1a
				0xf1,	// 0x1b
				0xa8,	// 0x1c
				0xf2,	// 0x1d
				0x89,	// 0x1e
				0x00,	// 0x1f
			};

			::memcpy(mReg, rw_reg, sizeof(rw_reg));

			mParam.channelFrequencykHz   = 0;
			mParam.crystalFrequencykHz   = 16000;
			mParam.fastSearchMode        = false;
			mParam.standby               = false;
			mParam.lowPassFilterWaitTime = 20;
			mParam.fastSearchWaitTime    = 4;
			mParam.normalSearchWaitTime  = 15;
		}

		status Init(Bus::I2C *bus)
		{
			// Dummy Access
			status	status;

			const uint8		INIT_DUMMY_RESET = 0x0C;

//			status = Write(bus, 0x01, INIT_DUMMY_RESET);
//			status = Write(bus, 0x01, INIT_DUMMY_RESET);

			// Soft Reset ON
			status = Write(bus, 0x01, INIT_DUMMY_RESET);
			if (status) return status;

			Sleep(bus, 1);

			// Soft Reset OFF
			uint8	i_data;

			i_data = mReg[0x01];

			i_data |= 0x10;
			status = Write(bus, 0x01 , i_data);
			if (status) return status;

			// ID Check
			status = Read(bus, 0x00, &i_data);
			if (status) return status;
			if (bus == NULL) {
				if (0x48 != i_data){
					return PT::STATUS_INVALID_PARAM_ERROR;
				}
			}

			// LPF Tuning On
			Sleep(bus, 1);
			mReg[0x0c] |= 0x40;
			status = Write(bus, 0x0c , mReg[0x0c]);
			if (status) return status;
			Sleep(bus, mParam.lowPassFilterWaitTime);

			for (uint8 i=0; i<0x20; i++){
				if (Flag(i) == 1){
					status = Write(bus, i , mReg[i]);
					if (status) return status;
				}
			}

			status = SetSleepMode(bus);
			if (status) return status;

			status = SetSearchMode(bus);
			if (status) return status;

			return status;
		}

		status SetSleep(bool sleep)
		{
			status	status;

			mParam.standby = sleep;

			if (sleep) {
				status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_MANUAL); CHECK_STATUS;
				SetSleepMode(I2C_NULL);
				mTC90522->SetSleep(I2C_NULL, sleep);
			} else {
				mTC90522->SetSleep(I2C_NULL, sleep);
				SetSleepMode(I2C_NULL);
			}

			mSleep = sleep;

			return PT::STATUS_OK;
		}

		void GetSleep(bool *sleep) const
		{
			_ASSERT(sleep);
			*sleep = mSleep;
		}

		status SetFrequency(uint32 channel, sint32 offset)
		{
			status	status;

			status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_MANUAL); CHECK_STATUS;

			bool bs;
			uint32 number, frequency;

			GetChannelFrequency(channel, &bs, &number, &frequency);

			uint32	frequencykHz = frequency*10 + offset;

			if (mTC90522->Index() == 0) {
				frequencykHz -= 500;
			} else {
				frequencykHz += 500;
			}
			mParam.channelFrequencykHz = frequencykHz;

			status = LocalLpfTuning(NULL); CHECK_STATUS;

			uint32	startTime;
			OS::Time::Get(&startTime);

			bool locked = false;
			while (true) {
				uint32	nowTime;
				OS::Time::Get(&nowTime);

				status = GetLocked(I2C_NULL, &locked);
				if (status) return status;
				if (locked) break;

				uint32	deltaTime = nowTime - startTime;
				if (100 <= deltaTime) break;

				OS::Thread::Sleep(1);
			}
			if (locked == false) return PT::STATUS_PLL_LOCK_TIMEOUT_ERROR;

			status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_AUTO); CHECK_STATUS;

			mChannel = channel;
			mOffset  = offset;

			return status;
		}

		void GetFrequency(uint32 *channel, sint32 *offset) const
		{
			_ASSERT(channel);
			*channel = mChannel;

			_ASSERT(offset);
			*offset = mOffset;
		}

		status GetLocked(Bus::I2C *bus, bool *locked)
		{
			_ASSERT(locked);
			status status = GetLockStatus(bus, locked);
			return status;
		}

	protected:
		friend class QM1D1C0042_;

		Param		mParam;
		uint8		mReg[0x20];

		WriteFunction	mWriteFunction;
		ReadFunction	mReadFunction;
		SleepFunction	mSleepFunction;
		void			*mFunctionArg;

		bool			mSleep;

		uint32			mChannel;
		sint32			mOffset;

		void SetReg(uint32 address, uint8 data)
		{
			mReg[address] = data;
		}

		void GetReg(uint32 address, uint8 *data)
		{
			*data = mReg[address];
		}

		uint8 Flag(uint32 index)
		{
			const uint8		FLAG[0x20] =	// 0:R, 1:R/W
			{	// _0 _1 _2 _3 _4 _5 _6 _7 _8 _9 _a _b _c _d _e _f
				  0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
				  0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
			};

			return FLAG[index];
		}

		status SetSleepMode(Bus::I2C *bus)
		{
			Param *param = &mParam;
			status status;

			if (param->standby) {
				mReg[0x01] &= (~(1 << 3)) & 0xff;	// BB_REG_disable
				mReg[0x01] |= 1 << 0;				// STDBY
				mReg[0x05] |= 1 << 3;				// pfd_rst STANDBY

				status = Write(bus, 0x05 , mReg[0x05]);
				if (status) return status;

				status = Write(bus, 0x01 , mReg[0x01]);
				if (status) return status;
			} else {
				mReg[0x01] |= 1 << 3;				// BB_REG_enable
				mReg[0x01] &= (~(1 << 0)) & 0xff;	// NORMAL
				mReg[0x05] &= (~(1 << 3)) & 0xff;	// pfd_rst NORMAL

				status = Write(bus, 0x01 , mReg[0x01]);
				if (status) return status;

				status = Write(bus, 0x05 , mReg[0x05]);
				if (status) return status;
			}

			return status;
		}

		status SetFastSearch(Bus::I2C *bus)
		{
			status status = SetSearchMode(bus);
			return status;
		}

		status SetFrequency(uint32 frequency)
		{
			_ASSERT(frequency);


			status status;

			mParam.channelFrequencykHz = frequency*10;

//			if (mLPF) {
				status = LocalLpfTuning(NULL, false);
//			} else {
//				status = LocalLpfTuning();
//				mLPF = true;
//			}
			return status;
		}

		status Tuning(Bus::I2C *bus, uint32 &sd)
		{
			Param	*param = &mParam;

			uint8 i_data;
			status status;

			// LPF
			mReg[0x08] &= 0xf0;
			mReg[0x08] |= 0x09;

			mReg[0x13] &= 0x9f;
			mReg[0x13] |= 0x20;

			const uint32 TABLE[9][3] = {
				{ 2151000, 1, 7 },
				{ 1950000, 1, 6 },
				{ 1800000, 1, 5 },
				{ 1600000, 1, 4 },
				{ 1450000, 1, 3 },
				{ 1250000, 1, 2 },
				{ 1200000, 0, 7 },
				{  975000, 0, 6 },
				{  950000, 0, 0 }
			};

			// div2/vco_band
			for (uint32 i=0; i<8; i++){
				if ((TABLE[i+1][0] <= param->channelFrequencykHz) && (param->channelFrequencykHz < TABLE[i][0])) {
					i_data = mReg[0x02];
					i_data &= 0x0f;
					i_data |= TABLE[i][1] << 7;
					i_data |= TABLE[i][2] << 4;
					status = Write(bus, 0x02, i_data);
				}
			}

			// PLL Counter
			_ASSERT(((int)(+1.4)) == +1);
			_ASSERT(((int)(+1.6)) == +1);
			_ASSERT(((int)(-1.4)) == -1);
			_ASSERT(((int)(-1.6)) == -1);

			double	M = static_cast<double>(param->channelFrequencykHz) / static_cast<double>(param->crystalFrequencykHz);
			uint32	a = static_cast<sint32>(M + 0.5);
			double	b = M - a;

			uint32	N = (a - 12) / 4;
			uint32	A = a - 4*(N + 1) - 5;

			if (0 <= b) {
				sd = static_cast<uint32>(pow(2, 20.)*b);
			} else {
				sd = static_cast<uint32>(pow(2, 20.)*b + (1<<22));
			}
			mReg[0x06] &= 0x40;
			mReg[0x06] |= N;
			status = Write(bus, 0x06 , mReg[0x06]);
			if (status) return status;

			mReg[0x07] &= 0xf0;
			mReg[0x07] |= A & 0x0f;
			status = Write(bus, 0x07 , mReg[0x07]);
			if (status) return status;

			return status;
		}

		status LocalLpfTuning(Bus::I2C *bus, bool lpf = true)
		{
			Param	*param = &mParam;
			uint8	i_data;
			uint32	sd;
			status	status;

			status = Tuning(bus, sd);
			if (status) return status;

			if (lpf) {
				// LPF_CLK , LPF_FC
				i_data = mReg[0x08] & 0xf0;
				i_data |= 2;
				status = Write(bus, 0x08 , i_data);
			} else {
				status = Write(bus, 0x08, mReg[0x08]);
				if (status) return status;
			}

			mReg[0x09] &= 0xc0;
			mReg[0x09] |= (sd >> 16) & 0x3f;
			mReg[0x0a]  = static_cast<uint8>(sd >> 8);
			mReg[0x0b]  = static_cast<uint8>(sd >> 0);
			status = Write(bus, 0x09 , mReg[0x09]);
			if (status) return status;

			status = Write(bus, 0x0a, mReg[0x0a]);
			if (status) return status;

			status = Write(bus, 0x0b, mReg[0x0b]);
			if (status) return status;

			if (lpf == false) {
				// CSEL_Offset
				status = Write(bus, 0x13, mReg[0x13]);
				if (status) return status;
			}

			if (lpf) {
				//. CO_TM , LPF_TM
				i_data = mReg[0x0c];
				i_data &= 0x3f;
				status = Write(bus, 0x0c , i_data);
				Sleep(bus, 1);	//1024usec

				// VCO_TM , LPF_TM
				i_data = mReg[0x0c];
				i_data |= 0xc0;
				status = Write(bus, 0x0c , i_data);
				if (status) return status;
			} else {
				// VCO_TM , LPF_TM
				i_data = mReg[0x0c];
				i_data &= 0x7f;
				status = Write(bus, 0x0c, i_data);
				if (status) return status;
				Sleep(bus, 1);	//1024usec

				// VCO_TM , LPF_TM
				i_data = mReg[0x0c];
				i_data |= 0x80;
				status = Write(bus, 0x0c, i_data);
				if (status) return status;
			}

			if (lpf) {
				Sleep(bus, param->lowPassFilterWaitTime);
			} else {
				if (mReg[0x03] & 0x01){		// Fast
					Sleep(bus, param->fastSearchWaitTime);
				}else{
					Sleep(bus, param->normalSearchWaitTime);
				}
			}

			if (lpf) {
				// LPF_FC
				status = Write(bus, 0x08, 0x09);
				if (status) return status;

				// CSEL_Offset
				status = Write(bus, 0x13, mReg[0x13]);
				if (status) return status;
			}

			// PLL Lock
			if (mReadFunction) {
				bool bLock;
				GetLockStatus(bus, &bLock);
				if(bLock != true){
					return PT::STATUS_PLL_LOCK_TIMEOUT_ERROR;
				}
			}
			return status;
		}

		status GetLockStatus(Bus::I2C *bus, bool* locked) 
		{
			_ASSERT(locked);

			status status = Read(bus, 0x0d, &mReg[0x0d]);
			if (status) return status;

			if (mReg[0x0d] & 0x40) {
				*locked = true;
			}else{
				*locked = false;
			}

			return status;
		}

		status SetSearchMode(Bus::I2C *bus) 
		{
			Param *param = &mParam;
			status status;

			if (param->fastSearchMode) {
				mReg[0x03] |= 0x01;
				status = Write(bus, 0x03 , mReg[0x03]);
				if (status) return status;
			} else {
				mReg[0x03] &= 0xfe;
				status = Write(bus, 0x03 , mReg[0x03]);
				if (status) return status;
			}
			return status;
		}

		status Write(Bus::I2C *bus, uint8 address, uint8 data)
		{
			if (mWriteFunction) {
				mWriteFunction(address, data, mFunctionArg);
				return 0;
			} else {
				status	status = mTC90522->WriteTuner(bus, address, &data, sizeof(data));
				mReg[address] = data;
				return status;
			}
		}

		status Read(Bus::I2C *bus, uint8 address, uint8 *data)
		{
			if (mReadFunction) {
				mReadFunction(address, data, mFunctionArg);
				return 0;
			} else {
				status status;
				if ((address ==  0x00) || (address == 0x0d)) {
					status = mTC90522->ReadTuner(bus, address, data, 1);
				} else {
					_ASSERT(false);
					status = 0;
				}
				return status;
			}
		}

		void Sleep(Bus::I2C *bus, uint32 ms)
		{
			if (mSleepFunction) {
				mSleepFunction(ms, mFunctionArg);
				return;
			} else {
				if (bus) {
					bus->Sleep(ms);
				} else {
					OS::Thread::Sleep(ms);
				}
			}
		}

	};
}

#endif
