#ifndef _TC90522_H
#define _TC90522_H

#include "Prefix.h"
#include "Bus_I2C.h"
#include "FPGA_I2C.h"
#include "EARTH_PT3.h"

namespace EARTH {
	class TC90522 {
	public:
		static uint8 Address(uint32 pin, PT::Device::ISDB isdb, uint32 index)
		{
			_ASSERT(pin <= BIT_2(1,1));
			_ASSERT(isdb < PT::Device::ISDB_COUNT);
			_ASSERT(index <= BIT_1(1)); 

			uint8	isdb2 = (isdb == PT::Device::ISDB_S) ? BIT_1(1) : BIT_1(0);

			uint8	address = static_cast<uint8>(
				  BIT_3(0,0,1) << 4
				| pin          << 2
				| index        << 1
				| isdb2        << 0);

			return address;
		}

		TC90522(FPGA::I2C *i2c, uint8 myAddress, uint8 tunerAddress)
			: mI2C(i2c), mMyAddress(myAddress), mTunerAddress(tunerAddress), mMasterClockFrequencyMHz(78)
		{
		}

		~TC90522()
		{
		}

		status Write(Bus::I2C *bus, uint8 address, const uint8 *data, uint32 size)
		{
			_ASSERT(data);
			_ASSERT(size);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(address);
			p->Write(data, size);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
			}

			return status;
		}

		status Write(Bus::I2C *bus, uint8 address, uint8 data)
		{
			status	status = Write(bus, address, &data, sizeof(data));
			return status;
		}

		status WriteTuner(Bus::I2C *bus, uint8 address, const uint8 *data, uint32 size)
		{
			_ASSERT(data);
			_ASSERT(size);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(THROUGH);
			p->Write(mTunerAddress << 1 | BIT_1(0));
			p->Write(address);
			p->Write(data, size);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
			}

			return status;
		}

		status WriteTunerWithoutAddress(Bus::I2C *bus, const uint8 *data, uint32 size)
		{
			_ASSERT(data);
			_ASSERT(size);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(THROUGH);
			p->Write(mTunerAddress << 1 | BIT_1(0));
			p->Write(data, size);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
			}

			return status;
		}

		status Read(Bus::I2C *bus, uint8 address, uint8 *readPtr, uint32 readSize = 1)
		{
			_ASSERT(readPtr);
			_ASSERT(readSize);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(address);

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(1));
			size_t	readIndex = p->Read(readSize);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
				for (uint32 i=0; i<readSize; i++) {
					readPtr[i] = p->Data1(readIndex+i);
				}
			}

			return status;
		}

		status ReadTuner(Bus::I2C *bus, uint8 address, uint8 *readPtr, uint32 readSize = 1)
		{
			_ASSERT(readPtr);
			_ASSERT(readSize);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(THROUGH);
			p->Write(mTunerAddress << 1 | BIT_1(0));
			p->Write(address);

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(THROUGH);
			p->Write(mTunerAddress << 1 | BIT_1(1));

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(1));
			size_t	readIndex = p->Read(readSize);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
				for (uint32 i=0; i<readSize; i++) {
					readPtr[i] = p->Data1(readIndex+i);
				}
			}

			return status;
		}

		status ReadTunerWithoutAddress(Bus::I2C *bus, uint8 *readPtr, uint32 readSize = 1)
		{
			_ASSERT(readPtr);
			_ASSERT(readSize);

			Bus::I2C	i2c;
			Bus::I2C	*p = bus ? bus : &i2c;

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(0));
			p->Write(THROUGH);
			p->Write(mTunerAddress << 1 | BIT_1(1));

			p->Start();
			p->Write(mMyAddress << 1 | BIT_1(1));
			size_t	readIndex = p->Read(readSize);
			p->Stop();

			status	status;

			if (bus) {
				status = PT::STATUS_OK;
			} else {
				p->End();
				status = mI2C->Run(p);
				for (uint32 i=0; i<readSize; i++) {
					readPtr[i] = p->Data1(readIndex+i);
				}
			}

			return status;
		}

		uint16 Byte2(const uint8 *ptr) const
		{
			return static_cast<uint16>(ByteN(ptr, 2)); 
		}

		uint32 Byte3(const uint8 *ptr) const
		{
			return ByteN(ptr, 3);
		}

		uint32 ByteN(const uint8 *ptr, uint32 n) const
		{
			_ASSERT(ptr);
			_ASSERT((1 <= n) && (n <= 4));

			uint32	value = 0;

			for (uint32 i=0; i<n; i++) {
				value <<= 8;
				value  |= ptr[i];
			}

			return value;
		}

		uint32 Index() const
		{
			uint32	index = BIT_SHIFT_MASK(mMyAddress, 1, 1);
			return index;
		}

		enum AGC {
			AGC_AUTO,
			AGC_MANUAL
		};

	protected:
		FPGA::I2C		*const	mI2C;
		const uint8		mMyAddress;
		const uint8		mTunerAddress;
		const double	mMasterClockFrequencyMHz;

		enum {
			THROUGH = 0xfe
		};
	};
}

#endif
