#ifndef _TC90522_S_H
#define _TC90522_S_H

#include "Prefix.h"
#include "FPGA_I2C.h"
#include "EARTH_PT3.h"
#include "TC90522.h"
#include <math.h>

namespace EARTH {
	class TC90522_S : public TC90522 {
	public:
		TC90522_S(FPGA::I2C *i2c, uint8 myAddress, uint8 tunerAddress)
			: TC90522(i2c, myAddress, tunerAddress)
		{
		}

		~TC90522_S()
		{
		}

		status Init(Bus::I2C *bus)
		{
			status	status;

			status = Write_pskmsrst(bus); CHECK_STATUS;
			status = Write(bus, 0x1e, 0x10); CHECK_STATUS;

			return status;
		}

		status SetTsPinsMode(Bus::I2C *bus, const PT::Device::TsPinsMode *mode)
		{
			uint32	clock_data = mode->clock_data,
					byte       = mode->byte,
					valid      = mode->valid;

			if (clock_data) clock_data++;
			if (byte      ) byte++;
			if (valid     ) valid++;

			uint8	data[2] = {
				static_cast<uint8>(0x15 | valid << 6),
				static_cast<uint8>(0x04 | clock_data << 4 | byte << 0)
			};

			status	status;

			status = Write(bus, 0x1c, data[0]); CHECK_STATUS;
			status = Write(bus, 0x1f, data[1]); CHECK_STATUS;

			return status;
		}

		status SetAGC(Bus::I2C *bus, AGC agc)
		{
			const uint8		DATA[2] = {
				0xb0, 0x30
			};

			status	status;

			uint8	data;

			data = (agc == AGC_AUTO) ? 0xff : 0x00;
			status = Write(bus, 0x0a, data); CHECK_STATUS;

			data  = DATA[Index()];
			data |= (agc == AGC_AUTO) ? 0x01 : 0x00;
			status = Write(bus, 0x10, data); CHECK_STATUS;

			data = (agc == AGC_AUTO) ? 0x40 : 0x00;
			status = Write(bus, 0x11, data); CHECK_STATUS;

			status = Write_pskmsrst(bus); CHECK_STATUS;

			return status;
		}

		status SetSleep(Bus::I2C *bus, bool sleep)
		{
			status	status;

			status = Write(bus, 0x17, sleep ? 1 : 0); CHECK_STATUS;

			return status;
		}

		status Write_Id(Bus::I2C *bus, uint16 id)
		{
			uint8 data[2] = { id >> 8, static_cast<uint8>(id) };

			status	status;

			status = Write(bus, 0x8f, data, sizeof(data)); CHECK_STATUS;

			return status;
		}

		status Write_jperst(Bus::I2C *bus)
		{
			status	status;

			status = Write(bus, 0xa5, 0xc0 | BIT_1(1) << 0); CHECK_STATUS;

			return status;
		}

		status Write_oponff(Bus::I2C *bus, bool null_h)
		{
			uint8	data = 0x04;
			if (null_h) {
				data |= BIT_1(1) << 3;
			}
			
			status	status;

			status = Write(bus, 0xa6, data); CHECK_STATUS;

			return status;
		}

		status Read_AGC(Bus::I2C *bus, uint8 *agc)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0xba, &data); CHECK_STATUS;

			_ASSERT(agc);
			*agc = data & 0x7f;

			return status;
		}

		status Read_afcfrq(Bus::I2C *bus, sint32 *hz)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0xbb, &data); CHECK_STATUS;

			sint8	afcfrq = static_cast<sint8>(data);

			double	value = afcfrq * mMasterClockFrequencyMHz * ::pow(10, static_cast<double>(6)) / 1024;
			if (0 <= value) {
				value += 0.5;
			} else {
				value -= 0.5;
			}
			
			_ASSERT(hz);
			*hz = static_cast<sint32>(value);

			return status;
		}

		status Read_CN(Bus::I2C *bus, uint32 *cn)
		{
			status	status;

			uint8	data[2];
			status = Read(bus, 0xbc, data, sizeof(data)); CHECK_STATUS;

			double	p = Byte2(data);

			p -= 3000;
			if (p < 0) {
				p = 0;
			}
			p = ::sqrt(p) / 64;
		
			double p2 = ::pow(p, 2);
			double p3 = ::pow(p, 3);
			double p4 = ::pow(p, 4);
			double p5 = ::pow(p, 5);

			p = -1.6346*p5 + 14.341*p4 - 50.259*p3 + 88.977*p2 - 89.565*p + 58.857;
			p *= 100;
			if (p < 0) {
				p = 0;
			}

			_ASSERT(cn);
			*cn = static_cast<uint32>(p + 0.5);

			return status;
		}

		status Read_clkfrq(Bus::I2C *bus, sint32 *ppm100)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0xbe, &data); CHECK_STATUS;

			sint8	clkfrq = static_cast<sint8>(data);

			double	value = clkfrq * 28.86 * 7.63 * 100 / mMasterClockFrequencyMHz;
			if (0 <= value) {
				value += 0.5;
			} else {
				value -= 0.5;
			}
			
			_ASSERT(ppm100);
			*ppm100 = static_cast<sint32>(value);

			return status;
		}

		status Read_tmcerr(Bus::I2C *bus, bool *tmcerr)
		{
			status	status;

			uint8 data;
			status = Read(bus, 0xc3, &data); CHECK_STATUS;

			_ASSERT(tmcerr);
			*tmcerr = BIT_SHIFT_MASK(data, 4, 1);

			return status;
		}

		status Read_TMCC(Bus::I2C *bus, PT::Device::TmccS *tmcc)
		{
			enum {
				BASE = 0xc5,
				SIZE = 0xe5 - BASE + 1
			};

			status	status;

			uint8	data[SIZE];
			status = Read(bus, 0xc3, data); CHECK_STATUS;
			if (BIT_SHIFT_MASK(data[0], 4, 1)) {
				return PT::STATUS_GENERAL_ERROR;
			}

			status = Read(bus, 0xce, data, 2); CHECK_STATUS;
			if (Byte2(data) == 0) {
				return PT::STATUS_GENERAL_ERROR;
			}

			status = Read(bus, 0xc3, data); CHECK_STATUS;
			tmcc->Emergency = BIT_SHIFT_MASK(data[0], 2, 1);
			tmcc->ExtFlag   = BIT_SHIFT_MASK(data[0], 1, 1);

			status = Read(bus, 0xc5, data, SIZE); CHECK_STATUS;
			tmcc->Indicator = BIT_SHIFT_MASK(data[0xc5 - BASE], 3, 5);
			tmcc->UpLink    = BIT_SHIFT_MASK(data[0xc7 - BASE], 0, 4);

			for (uint32 i=0; i<4; i++) {
				uint32	byteOffset = i / 2;
				uint32	bitOffset  = (i % 2) ? 0 : 4;
				tmcc->Mode[i] = BIT_SHIFT_MASK(data[0xc8+byteOffset - BASE], bitOffset, 4);
				tmcc->Slot[i] = BIT_SHIFT_MASK(data[0xca+i - BASE], 0, 6);
			}

			for (uint32 i=0; i<8; i++) {
				tmcc->Id[i] = Byte2(data + 0xce+i*2 - BASE);
			}

			return status;
		}

		status Read_tsid(Bus::I2C *bus, uint16 id[8])
		{
			// sizeof(id) ̓|C^̃TCYԂƂɒӁB

			status	status;

			uint8	data[2*8];
			for (uint8 i=0; i<sizeof(data); i+=4) {
				status = Read(bus, 0xce+i, data+i, 4); CHECK_STATUS;
			}

			const uint8		*p = data;
			for (uint32 i=0; i<8; i++) {
				id[i] = Byte2(p);
				p+= 2;
			}

			return status;
		}

		status Read_perr_pecy(Bus::I2C *bus, PT::Device::LayerIndex layerIndex, uint32 *numerator, uint32 *denominator)
		{
			uint8	perr[3];
			uint8	pecy[2];

			uint8	offset = (layerIndex == PT::Device::LAYER_INDEX_H) ? 0 : 1;
			offset *= 5;

			status	status;

			status = Read(bus, 0xeb+offset+0, perr, sizeof(perr)); CHECK_STATUS;
			status = Read(bus, 0xeb+offset+3, pecy, sizeof(pecy)); CHECK_STATUS;

			_ASSERT(numerator);
			*numerator = Byte3(perr);

			_ASSERT(denominator);
			*denominator = 8 * 204 * Byte2(pecy);

			return status;
		}

		status Read_Id(Bus::I2C *bus, uint16 *id)
		{
			status	status;

			uint8	data[2];
			status = Read(bus, 0xe6, data, sizeof(data)); CHECK_STATUS;

			_ASSERT(id);
			*id = Byte2(data);

			return status;
		}

		status Read_Layer(Bus::I2C *bus, PT::Device::LayerS *layer)
		{
			status	status;

			uint8	data[3];
			status = Read(bus, 0xe8, data, sizeof(data)); CHECK_STATUS;

			_ASSERT(layer);
			layer->Mode [0] = BIT_SHIFT_MASK(data[0], 0, 3);
			layer->Mode [1] = BIT_SHIFT_MASK(data[0], 4, 3);
			layer->Count[0] = BIT_SHIFT_MASK(data[2], 0, 6);
			layer->Count[1] = BIT_SHIFT_MASK(data[1], 0, 6);

			return status;
		}

		status Read_BeforeViterbi(Bus::I2C *bus, uint32 *count)
		{
			status	status;

			uint8	data[3];
			status = Read(bus, 0xf5, data, sizeof(data)); CHECK_STATUS;

			uint32	data3 = Byte3(data);

			_ASSERT(count);
			*count = data3;

			return status;
		}

	protected:
		status Write_pskmsrst(Bus::I2C *bus)
		{
			status	status;

			status = Write(bus, 0x03, BIT_1(1)); CHECK_STATUS;

			return status;
		}

	};
}

#endif
