#ifndef _PT_Device_H
#define _PT_Device_H

#include "EARTH_PT3.h"
#include "Prefix.h"
#include "FPGA.h"
#include "FPGA_I2C_.h"
#include "TC90522_S.h"
#include "TC90522_S_.h"
#include "TC90522_T.h"
#include "TC90522_T_.h"
#include "QM1D1C0042.h"
#include "QM1D1C0042_.h"
#include "MxL301RF.h"
#include "MxL301RF_.h"
#include "PT_Device_.h"
#include "PCI_Bus.h"
#include "PCI_Memory.h"
#include "Bus_I2C.h"
#include "Bus_I2C_.h"

#define I2C_NULL NULL

namespace EARTH {
namespace PCI    { class _; class Device; class Memory; }
namespace PT {

	class _Bus;
	class _Device;

	class _Device : public Device {
	public:
		_Device(void (*busFunction)(_Device *, void *), void *busArgument, PCI::Top *pci, const Bus::DeviceInfo *deviceInfo)
			: mBusFunction(busFunction), mBusArgument(busArgument), mPCI(pci), mDeviceInfo(*deviceInfo)
		{
			mDevice = NULL;
			mMemory = NULL;
			mFPGA = NULL;

			#if 2 <= PUBLIC_LEVEL
			mDevice_ = NULL;
			_ASSERT(mDevice_ == NULL);
			mDevice_ = new _Device_;
			_ASSERT(mDevice_);
			#endif

			for (uint32 i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner *tuner = &mTuner[i];
				_ASSERT(tuner);
				tuner->TC90522_S  = NULL;
				tuner->TC90522_T  = NULL;
				tuner->QM1D1C0042 = NULL;
				tuner->MxL301RF   = NULL;

				#if 3 <= PUBLIC_LEVEL
				tuner->TC90522_S_ = NULL;
				tuner->TC90522_T_ = NULL;
				tuner->MxL301RF_  = NULL;
				#endif
			}
		}

		~_Device()
		{
			#if 2 <= PUBLIC_LEVEL
			delete_(mDevice_);
			#endif

			Close();

			mBusFunction(this, mBusArgument);
		}

		virtual status Delete()
		{
			if (mDevice) return STATUS_DEVICE_MUST_BE_CLOSED_ERROR;

			delete this;

			return STATUS_OK;
		}

		#if 2 <= PUBLIC_LEVEL
		_Device_ *Device_() const
		{
			return mDevice_;
		}
		#endif

		virtual status Open()
		{
			if (mDevice) return STATUS_DEVICE_IS_ALREADY_OPEN_ERROR;

			status	status;

			PCI::Target		target;
			target.Bus      = mDeviceInfo.Bus;
			target.Slot     = mDeviceInfo.Slot;
			target.Function = mDeviceInfo.Function;

			#if 3 <= PUBLIC_LEVEL
			_ASSERT(mDevice_);
			status = mDevice_->CheckBaseAddress(mPCI, &mDeviceInfo);
			if (status) { Close(); return status; }
			#endif

			status = mPCI->NewDeviceMemory(&target, &mDevice, &mMemory);
			if (status) { Close(); return status; }

			uint8	configRevision;
			status = mDevice->ReadConfig(0x08, &configRevision, sizeof(configRevision));
			if (status) { Close(); return status; }

			if (configRevision != 01) { Close(); return STATUS_CONFIG_REVISION_ERROR; }

			FPGA::_::Argument	argument;
			argument.Pci = mDevice;

			_ASSERT(mFPGA == NULL);
			mFPGA = static_cast<FPGA ::_ *>(new FPGA::_(&argument));

			#if 2 <= PUBLIC_LEVEL
			_ASSERT(mDevice_);
			mDevice_->SetFPGA(mFPGA);
			#endif

			FPGA::Config	*config = mFPGA->Config();
			_ASSERT(config);

			if (mDeviceInfo.PTn != 3) { Close(); return STATUS_INVALID_PARAM_ERROR; }

			uint32	address;
			config->GetAddress(&address);
			if (address == 0) { Close(); return STATUS_PCI_BASE_ADDRESS_ERROR; }

			#if 3 <= PUBLIC_LEVEL
			status = mDevice_->CheckEnable();
			if (status) { Close(); return status; }
			#endif

			mFPGA->Register()->EnableWrite();
			mFPGA->Register()->EnableRead ();

			ReadConstantInfo();

			#if PUBLIC_LEVEL == 2
			bool	updateMode;
			uint8	minFpgaVersion,
					maxFpgaVersion;
			
			mDevice_->GetUpdateMode(&updateMode, &minFpgaVersion, &maxFpgaVersion);
			if (updateMode) {
				if ((minFpgaVersion <= mConstantInfo.Version_FPGA) && (mConstantInfo.Version_FPGA <= maxFpgaVersion)) {
					return STATUS_OK;
				} else {
					Close();
					return STATUS_FPGA_VERSION_ERROR;
				}
			}
			#endif

			if (mConstantInfo.Version_FPGA != 0x04) {
				Close();
				return STATUS_FPGA_VERSION_ERROR;
			}

			FPGA::I2C	*i2c = mFPGA->I2C();
			_ASSERT(i2c);

			for (uint32 i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);

				uint8	tcAddress, tunerAddress;
				uint32	pin = 0;

				tcAddress = TC90522::Address(pin, ISDB_S, i);
				tunerAddress = QM1D1C0042::Address(i);

				_ASSERT(tuner->TC90522_S == NULL);
				tuner->TC90522_S = new TC90522_S(i2c, tcAddress, tunerAddress);

				#if 3 <= PUBLIC_LEVEL
				_ASSERT(tuner->TC90522_S_ == NULL);
				tuner->TC90522_S_ = new TC90522_S_(i2c, tcAddress, tunerAddress);
				#endif

				tcAddress = TC90522::Address(pin, ISDB_T, i);
				tunerAddress = MxL301RF::Address(i);

				_ASSERT(tuner->TC90522_T == NULL);
				tuner->TC90522_T = new TC90522_T(i2c, tcAddress, tunerAddress);

				#if 3 <= PUBLIC_LEVEL
				_ASSERT(tuner->TC90522_T_ == NULL);
				tuner->TC90522_T_ = new TC90522_T_(i2c, tcAddress, tunerAddress);
				#endif

				_ASSERT(tuner->QM1D1C0042 == NULL);
				tuner->QM1D1C0042 = new QM1D1C0042(tuner->TC90522_S);

				_ASSERT(tuner->MxL301RF == NULL);
				tuner->MxL301RF = new MxL301RF(tuner->TC90522_T);

				#if 3 <= PUBLIC_LEVEL
				_ASSERT(tuner->MxL301RF_ == NULL);
				tuner->MxL301RF_ = new MxL301RF_(tuner->MxL301RF);
				#endif
				
				#if 3 <= PUBLIC_LEVEL
				if (i == 0) {
					QM1D1C0042_		dump(tuner->QM1D1C0042);
					dump.Dump();
				}

				if (i == 0) {
					tuner->MxL301RF_->Dump();
				}
				#endif
			}

//			uint32 dataWhenClose = mFPGA->DMA()->DataWhenClose();

//			status = mFPGA->Clock()->SetReset(dataWhenClose);
//			if (status) return status;

			status = mFPGA->Power()->SetLnbPowerWhenClose(Device::LNB_POWER_OFF);
			if (status) { Close(); return status; }

			i2c->Init();

			#if 3 <= PUBLIC_LEVEL
			FPGA::I2C_	i2c_(i2c);
//			i2c_.Test();
			#endif

			return status;
		}

		virtual status Close()
		{
			status	status;

			status = Status();
			if (status) return status;

			for (uint32 i=0; i<PT::Device::ISDB_COUNT; i++) {
				for (uint32 j=0; j<2; j++) {
					bool	b;
					GetTransferEnabled(static_cast<PT::Device::ISDB>(i), j, &b);
					if (b) {
						SetTransferEnabled(static_cast<PT::Device::ISDB>(i), j, false);
					}
				}
			}

			LnbPower	lnbPower;
			mFPGA->Power()->GetLnbPowerWhenClose(&lnbPower);
			mFPGA->Power()->SetLnbPower(lnbPower);

			mFPGA->I2C()->Reset();

			for (uint32 i=0; i<sizeof(mTuner)/sizeof(*mTuner); i++) {
				Tuner *tuner = &mTuner[i];
				_ASSERT(tuner);

				#if 3 <= PUBLIC_LEVEL
				delete_(tuner->TC90522_S_);
				delete_(tuner->TC90522_T_);
				delete_(tuner->MxL301RF_);
				#endif

				delete_(tuner->TC90522_S);
				delete_(tuner->TC90522_T);
				delete_(tuner->QM1D1C0042);
				delete_(tuner->MxL301RF);
			}

			#if 2 <= PUBLIC_LEVEL
			mDevice_->SetFPGA(NULL);
			#endif

			delete_(mFPGA);

			if (mDevice) {
				mDevice->Delete();
				mDevice = NULL;
			}

			if (mMemory) {
				mMemory->Delete();
				mMemory = NULL;
			}

			return status;
		}

		status ReadConstantInfo()
		{
			uint32	data[2];
			mFPGA->Register()->Read32(0x00, &data[0]);
			mFPGA->Register()->Read32(0x0c, &data[1]);

			ConstantInfo	*p = &mConstantInfo;
			p->PTn                          = BIT_SHIFT_MASK(data[0], 24, 8);
			p->Version_RegisterMap          = BIT_SHIFT_MASK(data[0], 16, 8);
			p->Version_FPGA                 = BIT_SHIFT_MASK(data[0],  8, 8);
			p->CanTransportTS               = BIT_SHIFT_MASK(data[1],  5, 1) ? true : false;
			p->BitLength_PageDescriptorSize = BIT_SHIFT_MASK(data[1],  0, 5);

			return STATUS_OK;
		}

		virtual status GetConstantInfo(ConstantInfo *info) const
		{
			if (info == NULL) return STATUS_INVALID_PARAM_ERROR;

			*info = mConstantInfo;

			return STATUS_OK;
		}

		virtual status SetLnbPower(LnbPower lnbPower)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mFPGA->Power()->SetLnbPower(lnbPower);
	
			return status;
		}

		virtual status GetLnbPower(LnbPower *lnbPower) const
		{
			status status;

			status = Status();
			if (status) return status;

			if (lnbPower == NULL) return STATUS_INVALID_PARAM_ERROR;

			status = mFPGA->Power()->GetLnbPower(lnbPower);

			return status;
		}

		virtual status SetLnbPowerWhenClose(LnbPower lnbPower)
		{
			status status;

			status = Status();
			if (status) return status;

			status = mFPGA->Power()->SetLnbPowerWhenClose(lnbPower);

			return status;
		}

		virtual status GetLnbPowerWhenClose(LnbPower *lnbPower) const
		{
			status status;

			status = Status();
			if (status) return status;

			if (lnbPower == NULL) return STATUS_INVALID_PARAM_ERROR;

			status = mFPGA->Power()->GetLnbPowerWhenClose(lnbPower);

			return status;
		}

		virtual status InitTuner()
		{
			status	status = 0;

			bool useROM = true/*true*/;

			#if PUBLIC_LEVEL < 3
			useROM = true;
			#endif

			if (useROM == false) {
				status = CheckDefaultData();
			}

			for (uint32 i=0; i<2; i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);
			}

			EARTH::Bus::I2C		bus;
			uint32	startAddress[2];

			if (useROM) {
				bus.End();
				startAddress[0] = 4096 + (   0);
				startAddress[1] = 4096 + (2042);
			} else {
				startAddress[0] = 0;
				startAddress[1] = 0;

				#if 3 <= PUBLIC_LEVEL
				MakeROM(&bus, &startAddress[1]);
				#endif
				bus.End();

				::wprintf(L"StartAddress[1]:%d InstCount:%d\n", startAddress[1], bus.InstructionCount());

				#if 3 <= PUBLIC_LEVEL
				EARTH::Bus::I2C_	bus_(&bus);
				bus_.Dump();
				#endif

				mFPGA->I2C()->Copy(&bus);
			}

			bus.SetInstructionAddress(startAddress[0]);

			bool	clean;
			mFPGA->I2C()->GetClean(&clean);
			if (clean) {
				//::wprintf(L"Clean tOZbgĂ邽߁AIC ̏XLbv܂B\n");
			} else {
				status = mFPGA->I2C()->Run(&bus, NULL, false);
				if (status) return status;
			}

			status = CheckWriteData();
			if (status) return status;

			status = TunerPowerOn(&bus, startAddress[1]);
			if (status) return status;

			for (uint32 i=0; i<ISDB_COUNT; i++) {
				ISDB	isdb = static_cast<ISDB>(i);
				for (uint32 j=0; j<2; j++) {
					status = SetTunerSleep(isdb, j, false);
					if (status) return status;

					uint32	channel;
					if (isdb == ISDB_S) {
						channel = 0;
					} else {
						channel = (j == 0) ? 70 : 71;
					}
					status = SetFrequency(isdb, j, channel);
					if (status) return status;
				}
			}

			for (uint32 i=0; i<ISDB_COUNT; i++) {
				ISDB	isdb = static_cast<ISDB>(i);
				for (uint32 j=0; j<2; j++) {
					status = SetTunerSleep(isdb, j, true);
					if (status) return status;
				}
			}

			return status;
		}

		status CheckDefaultData()
		{
			status	status = 0;

			// Reset
			EARTH::Bus::I2C		bus;
			bus.Reset();
			bus.End();
			mFPGA->I2C()->Run(&bus, NULL);

			for (uint32 i=0; i<2; i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);

				#if 3 <= PUBLIC_LEVEL
				status = tuner->TC90522_S_->CheckDefaultData(NULL);
				if (status) return status;

				status = tuner->TC90522_T_->CheckDefaultData(NULL);
				if (status) return status;
				#endif
			}

			return status;
		}

		status CheckWriteData()
		{
			status	status = 0;

			for (uint32 i=0; i<2; i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);

				#if 3 <= PUBLIC_LEVEL
				status = tuner->TC90522_S_->CheckWriteData(NULL);
				if (status) return status;

				status = tuner->TC90522_T_->CheckWriteData(NULL);
				if (status) return status;
				#endif
			}

			return status;
		}

		virtual status SetAmpPowerT(bool b)
		{
			Tuner	*tuner = &mTuner[1];

			tuner->TC90522_T->SetAmpPower(b);
			status	status = tuner->TC90522_T->SetPowers(I2C_NULL);
			if (status) return status;
			
			return status;
		}

		virtual status SetTunerSleep(ISDB isdb, uint32 tunerIndex, bool sleep)
		{
			status	status;

			status = Status(isdb, tunerIndex, false);
			if (status) return status;

			Tuner	*tuner = &mTuner[tunerIndex];
			_ASSERT(tuner);

			switch (isdb) {
				case ISDB_S: status = tuner->QM1D1C0042->SetSleep(sleep); break;
				case ISDB_T: status = tuner->MxL301RF  ->SetSleep(sleep); break;
			}
	
			return status;

		}

		virtual status GetTunerSleep(ISDB isdb, uint32 tunerIndex, bool *sleep) const
		{
			status status;

			status = Status(isdb, tunerIndex, false);
			if (status) return status;

			if (sleep == NULL) return STATUS_INVALID_PARAM_ERROR;

			const Tuner		*tuner = &mTuner[tunerIndex];
			_ASSERT(tuner);

			switch (isdb) {
				case ISDB_S: tuner->QM1D1C0042->GetSleep(sleep); break;
				case ISDB_T: tuner->MxL301RF  ->GetSleep(sleep); break;
			}

			return status;
		}

		virtual status SetFrequency(ISDB isdb, uint32 tunerIndex, uint32 channel, sint32 offset = 0)
		{
			status	status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			Tuner	*tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S: status = tuner->QM1D1C0042->SetFrequency(channel, offset); break;
				case ISDB_T: status = tuner->MxL301RF  ->SetFrequency(channel, offset); break;
			}

			return status;
		}

		virtual status GetFrequency(ISDB isdb, uint32 tunerIndex, uint32 *channel, sint32 *offset) const
		{
			status	status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			if ((channel == NULL) || (offset == NULL)) return STATUS_INVALID_PARAM_ERROR;

			const Tuner		*tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S: tuner->QM1D1C0042->GetFrequency(channel, offset); break;
				case ISDB_T: tuner->MxL301RF  ->GetFrequency(channel, offset); break;
			}

			return status;
		}

		virtual status GetFrequencyOffset(ISDB isdb, uint32 tunerIndex, sint32 *clock, sint32 *carrier)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			if ((clock == NULL) || (carrier == NULL)) return STATUS_INVALID_PARAM_ERROR;

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90522_S->Read_clkfrq(I2C_NULL, clock);
					tuner->TC90522_S->Read_afcfrq(I2C_NULL, carrier);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90522_T->Read_clkafc_dt(I2C_NULL, clock);
					tuner->TC90522_T->Read_carafc_dt(I2C_NULL, carrier);
					break;
				}
			}

			return status;
		}

		virtual status GetCnAgc(ISDB isdb, uint32 tunerIndex, uint32 *cn100, uint32 *currentAgc, uint32 *maxAgc)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			if ((cn100 == NULL) || (currentAgc == NULL) || (maxAgc == NULL)) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90522_S->Read_CN(I2C_NULL, cn100);

					uint8 byteAgc;
					tuner->TC90522_S->Read_AGC(I2C_NULL, &byteAgc);
					_ASSERT(currentAgc); *currentAgc = byteAgc;
					_ASSERT(maxAgc); *maxAgc = 127;
					break;
				}
				case ISDB_T:
				{
					tuner->TC90522_T->Read_cndat(I2C_NULL, cn100);

					uint8 byteAgc;
					tuner->TC90522_T->Read_ifagc_dt(I2C_NULL, &byteAgc);
					_ASSERT(currentAgc); *currentAgc = byteAgc;
					_ASSERT(maxAgc); *maxAgc = 255;
					break;
				}
			}

			return status;
		}

		virtual status GetRFLevel(uint32 tunerIndex, float *level)
		{
			status status;

			status = Status(ISDB_T, tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			tuner->MxL301RF->GetLevel(I2C_NULL, level);

			return status;
		}

		virtual status SetIdS(uint32 tunerIndex, uint32 id)
		{
			status status;

			status = Status(ISDB_S, tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];
			tuner->TC90522_S->Write_Id(I2C_NULL, static_cast<uint16>(id));

			return status;
		}

		virtual status GetIdS(uint32 tunerIndex, uint32 *id)
		{
			status status;

			status = Status(ISDB_S, tunerIndex);
			if (status) return status;

			if (id == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			uint16 shortId;
			tuner->TC90522_S->Read_Id(I2C_NULL, &shortId);
			
			*id = shortId;

			return STATUS_OK;
		}

		virtual status SetInnerErrorRateLayer(ISDB isdb, uint32 tunerIndex, LayerIndex /*layerIndex*/)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			return STATUS_GENERAL_ERROR;
		}

		virtual status GetInnerErrorRate(ISDB isdb, uint32 tunerIndex, ErrorRate *errorRate)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			if (errorRate == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90522_S->Read_BeforeViterbi(I2C_NULL, &errorRate->Numerator);
					errorRate->Denominator = 8*1632*48*8*2;
					break;
				}
				case ISDB_T:
				{
					tuner->TC90522_T->Read_BeforeViterbi(I2C_NULL, &errorRate->Numerator);
					errorRate->Denominator = 2*1632*4096*32;
					break;
				}
			}

			return status;
		}

		virtual status GetCorrectedErrorRate(ISDB isdb, uint32 tunerIndex, LayerIndex layerIndex, ErrorRate *errorRate)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			if (errorRate == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					if (LAYER_COUNT_S <= layerIndex) return STATUS_INVALID_PARAM_ERROR;
					tuner->TC90522_S->Read_perr_pecy(I2C_NULL, layerIndex, &errorRate->Numerator, &errorRate->Denominator);
					break;
				}
				case ISDB_T:
				{
					if (LAYER_COUNT_T <= layerIndex) return STATUS_INVALID_PARAM_ERROR;
					tuner->TC90522_T->Read_perr_pecy(I2C_NULL, layerIndex, &errorRate->Numerator, &errorRate->Denominator);
					break;
				}
			}

			return status;
		}

		virtual status ResetCorrectedErrorCount(ISDB isdb, uint32 tunerIndex)
		{
			status	status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			Tuner	*tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90522_S->Write_jperst(I2C_NULL);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90522_T->Write_perst(I2C_NULL);
					break;
				}
			}

			return STATUS_OK;
		}

		virtual status GetErrorCount(ISDB isdb, uint32 tunerIndex, uint32 *count)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			if (count == NULL) return STATUS_INVALID_PARAM_ERROR;

			mFPGA->DMA()->GetTsError(isdb*2 + tunerIndex, count);

			return status;
		}

		virtual status GetTmccS(uint32 tunerIndex, TmccS *tmcc)
		{
			status status;

			status = Status(ISDB_S, tunerIndex);
			if (status) return status;

			if (tmcc == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];
			status = tuner->TC90522_S->Read_TMCC(I2C_NULL, tmcc);

			return status;
		}

		virtual status GetLayerS(uint32 tunerIndex, LayerS *layer)
		{
			status status;

			status = Status(ISDB_S, tunerIndex);
			if (status) return status;

			if (layer == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];
			tuner->TC90522_S->Read_Layer(I2C_NULL, layer);

			return status;
		}

		virtual status GetTmccT(uint32 tunerIndex, TmccT *tmcc)
		{
			status status;

			status = Status(ISDB_T, tunerIndex);
			if (status) return status;

			if (tmcc == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			bool b = false;
			while (true) {
				bool	retryov;
				bool	tmunvld;
				bool	fulock;
				tuner->TC90522_T->Read_retryov_tmunvld_fulock(I2C_NULL, &retryov, &tmunvld, &fulock);
				if (!fulock) {
					b = true;
					break;
				} else {
					if (retryov) {
						break;
					}
				}
				OS::Thread::Sleep(1);
			}

			if (b) {
				tuner->TC90522_T->Read_TMCC(I2C_NULL, tmcc);
			}

			return b ? STATUS_OK : STATUS_GENERAL_ERROR;
		}
/*
		virtual status GetLockedT(uint32 tunerIndex, bool locked[3])
		{
			status status;

			status = Status(ISDB_T, tunerIndex);
			if (status) return status;

			if (locked == NULL) return STATUS_INVALID_PARAM_ERROR;

			Tuner *tuner = &mTuner[tunerIndex];

			bool rlocka, rlockb, rlockc;
			tuner->TC90522_T->Read_rlock(I2C_NULL, &rlocka, &rlockb, &rlockc);

			_ASSERT(locked);
			locked[0] = rlocka;
			locked[1] = rlockb;
			locked[2] = rlockc;

			if (rlockb) {
				uint32 ffsize, gdleng, seqen;
				tuner->TC90522_T->Read_ffsize_gdleng_seqen(I2C_NULL, &ffsize, &gdleng, &seqen);
				_ASSERT(8 <= seqen);
			}

			return status;
		}
*/
		virtual status SetLayerEnable(ISDB isdb, uint32 tunerIndex, LayerMask layerMask)
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			Tuner *tuner = &mTuner[tunerIndex];

			switch (isdb) {
				case ISDB_S:
				{
					if ((LAYER_MASK_L | LAYER_MASK_H) < layerMask) return STATUS_INVALID_PARAM_ERROR;
					bool null_h = !(layerMask & LAYER_MASK_H);
					tuner->TC90522_S->Write_oponff(I2C_NULL, null_h);
					break;
				}
				case ISDB_T:
				{
					if ((LAYER_MASK_A | LAYER_MASK_B | LAYER_MASK_C) < layerMask) return STATUS_INVALID_PARAM_ERROR;
					tuner->TC90522_T->Write_laysel(I2C_NULL, static_cast<uint8>(layerMask));
					break;
				}
			}

			tuner->LayerMask[isdb] = layerMask;

			return status;
		}

		virtual status GetLayerEnable(ISDB isdb, uint32 tunerIndex, LayerMask *layerMask) const
		{
			status status;

			status = Status(isdb, tunerIndex);
			if (status) return status;

			const Tuner *tuner = &mTuner[tunerIndex];

			if (layerMask == NULL) return PT::STATUS_INVALID_PARAM_ERROR;

			*layerMask = tuner->LayerMask[isdb];

			return PT::STATUS_OK;
		}

		// TS s
		virtual status SetTsPinsMode(ISDB isdb, uint32 index, const TsPinsMode *mode)
		{
			status status;

			status = Status(isdb, index);
			if (status) return status;

			Tuner *tuner = &mTuner[index];

			switch (isdb) {
				case ISDB_S:
				{
					tuner->TC90522_S->SetTsPinsMode(I2C_NULL, mode);
					break;
				}
				case ISDB_T:
				{
					tuner->TC90522_T->SetTsPinsMode(I2C_NULL, mode);
					break;
				}
			}

			return STATUS_OK;
		}

		virtual status GetTsPinsLevel(ISDB isdb, uint32 index, TsPinsLevel *level)
		{
			status status;

			status = Status(isdb, index);
			if (status) return status;

			mFPGA->DMA()->GetTsPinsLevel(isdb*2 + index, level);

			return STATUS_OK;
		}

		virtual status GetTsSyncByte(ISDB isdb, uint32 index, uint8 *syncByte)
		{
			status status;

			status = Status(isdb, index);
			if (status) return status;

			mFPGA->DMA()->GetTsSyncByte(isdb*2 + index, syncByte);

			return STATUS_OK;
		}

		virtual status SetRamPinsMode(RamPinsMode mode)
		{
			uint32	data;
			switch (mode) {
				case RAM_PINS_MODE_NORMAL: data = BIT_2(0, 0); break;
				case RAM_PINS_MODE_LOW   : data = BIT_2(0, 1); break;
				case RAM_PINS_MODE_HIGH  : data = BIT_2(1, 1); break;
				default                  : data = BIT_2(0, 0); break;
			}

			mFPGA->Register()->Write32(0x18, data);

			return STATUS_OK;
		}

		// DMA obt@
		virtual status LockBuffer__Obsolated__(void *ptr, uint32 size, void **handle)
		{
			uint32	d = PCI::Memory::DIRECTION_WRITE | PCI::Memory::DIRECTION_READ;
			status	status = mMemory->LockBuffer(ptr, size, static_cast<PCI::Memory::Direction>(d), handle);
			return status;
		}

		virtual status LockBuffer(void *ptr, uint32 size, TransferDirection direction, void **handle)
		{
			uint32	d = 0;
			if (direction & TRANSFER_DIRECTION_WRITE) d |= PCI::Memory::DIRECTION_WRITE;
			if (direction & TRANSFER_DIRECTION_READ ) d |= PCI::Memory::DIRECTION_READ;

			status	status = mMemory->LockBuffer(ptr, size, static_cast<PCI::Memory::Direction>(d), handle);
			return status;
		}

		virtual status SyncBufferCpu(void *handle)
		{
			return mMemory->SyncCpu(handle);
		}

		virtual status SyncBufferIo(void *handle)
		{
			return mMemory->SyncIo(handle);
		}

		virtual status UnlockBuffer(void *handle)
		{
			status	status = mMemory->UnlockBuffer(handle);
			return status;
		}

		virtual status GetBufferInfo(void *handle, const BufferInfo **infoTable, uint32 *infoCount)
		{
			status	status = mMemory->GetBufferInfo(handle, infoTable, infoCount);
			return status;
		}

		// DMA
		virtual status SetTransferPageDescriptorAddress(ISDB isdb, uint32 tunerIndex, uint64 pageDescriptorAddress)
		{
			_ASSERT(mFPGA);
			mFPGA->DMA()->SetAddress(isdb*2+tunerIndex, pageDescriptorAddress);

			return STATUS_OK;
		}

		virtual status SetTransferEnabled(ISDB isdb, uint32 tunerIndex, bool  enabled)
		{
			_ASSERT(mFPGA);
			mFPGA->DMA()->SetEnabled(isdb*2+tunerIndex, enabled);

			return STATUS_OK;
		}

		virtual status GetTransferEnabled(ISDB isdb, uint32 tunerIndex, bool *enabled) const
		{
			_ASSERT(mFPGA);
			mFPGA->DMA()->GetEnabled(isdb*2+tunerIndex, enabled);

			return STATUS_OK;
		}
		
		virtual status SetTransferTestMode(ISDB isdb, uint32 tunerIndex, bool testMode = false, uint16 initial = 0, bool not = false/*, bool resetError = false*/)
		{
			_ASSERT(mFPGA);
			mFPGA->DMA()->SetTestMode(isdb*2+tunerIndex, testMode, initial, not, false/*resetError*/);
			return 0;
		}

		virtual status GetTransferInfo(ISDB isdb, uint32 tunerIndex, TransferInfo *transferInfo)
		{
			mFPGA->DMA()->GetTransferInfo(isdb*2+tunerIndex, transferInfo);

			return STATUS_OK;
		}

	protected:
		void (*const mBusFunction)(_Device *, void *);
		void					*const mBusArgument;
		PCI::Top *const			mPCI;
		const Bus::DeviceInfo	mDeviceInfo;

		PCI::Device				*mDevice;
		PCI::Memory				*mMemory;
		FPGA::_					*mFPGA;
		ConstantInfo			mConstantInfo;

		#if 2 <= PUBLIC_LEVEL
		_Device_				*mDevice_;
		#endif

		struct Tuner {
			TC90522_S	*TC90522_S;
			TC90522_T	*TC90522_T;
			QM1D1C0042	*QM1D1C0042;
			MxL301RF	*MxL301RF;

			#if 3 <= PUBLIC_LEVEL
			TC90522_S_	*TC90522_S_;
			TC90522_T_	*TC90522_T_;
			MxL301RF_	*MxL301RF_;
			#endif

			LayerMask	LayerMask[2];
		};

		Tuner	mTuner[2];

		status TunerPowerOn(EARTH::Bus::I2C *bus, uint32 startAddress)
		{
			status	status;

			Tuner	*tuner;

			for (uint32 i=0; i<2; i++) {
				tuner = &mTuner[i];
				tuner->TC90522_S->Init(I2C_NULL);
			}
			
			for (uint32 i=0; i<2; i++) {
				tuner = &mTuner[i];
				tuner->TC90522_T->Init(I2C_NULL);
			}

			OS::Thread::Sleep(1);

			tuner = &mTuner[1];
			tuner->TC90522_T->SetTunerPower(true);
			tuner->TC90522_T->SetAmpPower(false);
			status = tuner->TC90522_T->SetPowers(I2C_NULL);
			if (status) return status;

			// dオĈ肷܂ŎԂ|̂ŁA TS sm[}ԂɂĂ
			PT::Device::TsPinsMode	pins;

			pins.clock_data = TS_PIN_MODE_NORMAL;
			pins.byte       = TS_PIN_MODE_NORMAL;
			pins.valid      = TS_PIN_MODE_NORMAL;

			for (uint32 i=0; i<2; i++) {
				tuner = &mTuner[i];
				tuner->TC90522_S->SetTsPinsMode(I2C_NULL, &pins);
			}

			for (uint32 i=0; i<2; i++) {
				tuner = &mTuner[i];
				tuner->TC90522_T->SetTsPinsMode(I2C_NULL, &pins);
			}

			OS::Thread::Sleep(1);

			status = InitTuner(0);
			if (status) return status;

			status = InitTuner(1);
			if (status) return status;

			if (startAddress < 4096) {
				mFPGA->I2C()->Copy(bus);
			}
			bus->SetInstructionAddress(startAddress);

			status = mFPGA->I2C()->Run(bus, NULL, false);
			if (status) return status;

			tuner->TC90522_T->SetAmpPower(true);
			status = tuner->TC90522_T->SetPowers(I2C_NULL);
			if (status) return status;

			return status;
		}

		status TunerPowerOff()
		{
			return 0;
		}

		status InitTuner(uint32 tunerIndex)
		{
			status	status;

//			status = Status(tunerIndex);
//			if (status) return status;

			Tuner	*tuner = &mTuner[tunerIndex];
			_ASSERT(tuner);

			tuner->QM1D1C0042->InitRegParam();

			{
				EARTH::Bus::I2C		bus;
				tuner->QM1D1C0042->DummyReset(&bus);
				bus.End();
				mFPGA->I2C()->Run(&bus);
			}

			{
				EARTH::Bus::I2C		bus;
				status = tuner->QM1D1C0042->Init(&bus);
				if (status) return status;
				bus.End();
				mFPGA->I2C()->Run(&bus);
			}

			return status;
		}

		#if 3 <= PUBLIC_LEVEL
		status MakeROM(EARTH::Bus::I2C *bus, uint32 *startAddress)
		{
			status	status = 0;

			_ASSERT(bus);
			bus->Reset();

			for (uint32 i=0; i<2; i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);

				status = tuner->TC90522_S_->Init(bus);
				if (status) return status;

				status = tuner->TC90522_T_->Init(bus);
				if (status) return status;
			}

			bus->End();
			_ASSERT(bus->ReadAddress() == 0);

			*startAddress = bus->InstructionCount();

			for (uint32 i=0; i<2; i++) {
				Tuner	*tuner = &mTuner[i];
				_ASSERT(tuner);

				status = tuner->MxL301RF_->Init(bus, true);
				if (status) return status;
			}
			_ASSERT(bus->ReadAddress() == 2);

			return status;
		}
#endif

		status Status() const
		{
			if (mDevice == NULL) return STATUS_DEVICE_IS_NOT_OPEN_ERROR;
			return STATUS_OK;
		}

		status Status(ISDB isdb, uint32 tunerIndex, bool checkSleep = true) const
		{
			if (mDevice == NULL) return STATUS_DEVICE_IS_NOT_OPEN_ERROR;
			if ((2 <= tunerIndex) || (ISDB_COUNT <= isdb)) return STATUS_INVALID_PARAM_ERROR;

			if (checkSleep) {
				const Tuner		*tuner = &mTuner[tunerIndex];
				_ASSERT(tuner);

				bool	sleep = true;
				switch (isdb) {
					case ISDB_S: tuner->QM1D1C0042->GetSleep(&sleep); break;
					case ISDB_T: tuner->MxL301RF  ->GetSleep(&sleep); break;
				}

				if (sleep) return STATUS_TUNER_IS_SLEEP_ERROR;
			}

			return STATUS_OK;
		}
	};
}
}

#endif
